from dataclasses import dataclass

from dbt.adapters.bigquery.connections import (  # type:ignore
    BigQueryConnectionManager,
    BigQueryCredentials,
)

from common.credentials import LayerCredentials


@dataclass
class LayerBigQueryCredentials(LayerCredentials, BigQueryCredentials):
    @property
    def type(self) -> str:
        return "layer_bigquery"


class LayerBigQueryConnectionManager(BigQueryConnectionManager):
    TYPE = "layer_bigquery"



import decimal
import pathlib
from typing import Any, List, Mapping

import agate  # type:ignore
import numpy as np
import pandas as pd  # type:ignore
from dbt.clients import agate_helper  # type:ignore


def from_agate_table(table: agate.Table, column_names_map: Mapping[str, str]) -> pd.DataFrame:
    """
    Converts the given agate table to a pandas dataframe
    """
    column_names = [column_names_map.get(column.upper(), column) for column in table.column_names]

    if len(table.rows) > 0:
        column_types = [_type_from_value(val) for val in table.rows[0]]
        rows = [_from_agate_row(row, column_types) for row in table.rows]
    else:
        rows = []
    return pd.DataFrame.from_records(rows, columns=column_names)


def _type_from_value(value: Any) -> Any:
    """
    Returns a function to cast the value to a suitable pandas data type
    """
    if isinstance(value, decimal.Decimal):
        dec_tuple = value.as_tuple()
        if dec_tuple.exponent >= 0:
            return np.int64
        return np.float64
    return None


def _from_agate_row(row: agate.Row, column_types: List[Any]) -> List[Any]:
    """
    Converts a given agate row to a plain python row
    """
    return [column_type(val) if column_type else val for val, column_type in zip(row, column_types)]


def _only_whole_numbers(series: pd.Series) -> bool:
    return series.apply(lambda x: x.is_integer()).all()


def _dataframe_to_csv(df: pd.DataFrame, path: pathlib.Path) -> None:
    """
    Type conversion necessary for the edge case where the pandas column is of float type but no values have a fraction.
    In this case the BQ adapter will infer from the agate table object the BQ column type to be an integer
    but it will fail to parse the csv containing float numbers.
    """
    for column in df.columns:
        if np.dtype(np.float64) == df.dtypes[column] and _only_whole_numbers(df[column]):
            df = df.astype({column: np.int64})
    df.to_csv(path, index=False)


def to_agate_table_with_path(dataframe: pd.DataFrame, path: pathlib.Path) -> agate.Table:
    """
    Converts the given pandas dataframe to an agate table
    """
    _dataframe_to_csv(dataframe, path)
    table = agate_helper.from_csv(path, [])
    table.original_abspath = path
    return table



from dbt.adapters.base import AdapterPlugin  # type: ignore

from dbt.adapters.layer_snowflake.connections import LayerSnowflakeCredentials
from dbt.adapters.layer_snowflake.impl import LayerSnowflakeAdapter
from dbt.include import layer_snowflake


Plugin = AdapterPlugin(
    adapter=LayerSnowflakeAdapter,
    credentials=LayerSnowflakeCredentials,
    include_path=layer_snowflake.PACKAGE_PATH,
    dependencies=["snowflake"],
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['common',
 'common.automl_models',
 'dbt',
 'dbt.adapters.layer_bigquery',
 'dbt.adapters.layer_snowflake',
 'dbt.include.layer_bigquery',
 'dbt.include.layer_snowflake']

package_data = \
{'': ['*']}

install_requires = \
['dbt-core==1.2.0',
 'layer==0.10.3084138152',
 'matplotlib==3.5.1',
 'pandas==1.3.5',
 'scikit-learn==1.0.2',
 'sqlparse>=0.4.2,<0.5.0',
 'typing-extensions>=4.3.0',
 'xgboost==1.5.1']

extras_require = \
{'bigquery': ['dbt-bigquery==1.2.0'], 'snowflake': ['dbt-snowflake==1.2.0']}

setup_kwargs = {
    'name': 'dbt-layer',
    'version': '0.1.3086121632',
    'description': 'The Layer adapter plugin for dbt',
    'long_description': '# Layer dbt Adapter for BigQuery\n\nThis adapter runs dbt builds for ML pipelines with BigQuery as the backing data warehouse.\n',
    'author': 'Layer',
    'author_email': 'info@layer.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<3.11',
}


setup(**setup_kwargs)



from typing import Any, List, Type

import pandas as pd  # type: ignore

import layer
from common.automl_models.base_model import AutoMLModel, TrainDataset
from common.automl_models.sklearn_models import (
    ScikitLearnAdaBoostClassifier,
    ScikitLearnDecisionTreeClassifier,
    ScikitLearnDecisionTreeRegressor,
    ScikitLearnLinearRegression,
    ScikitLearnRandomForestClassifier,
    ScikitLearnRidgeClassifier,
)
from common.automl_models.xgboost_models import XGBoostClassifier, XGBoostRegressor
from layer.decorators import model as model_decorator


class AutoML:
    automl_models: List[Type[AutoMLModel]] = [
        # sklearn models
        ScikitLearnRandomForestClassifier,
        ScikitLearnDecisionTreeClassifier,
        ScikitLearnAdaBoostClassifier,
        ScikitLearnRidgeClassifier,
        ScikitLearnLinearRegression,
        ScikitLearnDecisionTreeRegressor,
        # xgboost models
        XGBoostClassifier,
        XGBoostRegressor,
    ]

    def __init__(self, model_type: str, df: pd.DataFrame, features: List[str], target: str) -> None:
        self.model_type = model_type
        self.df = df
        self.features = features
        self.target = target
        self.score = None

    def train(self, project_name: str, model_name: str) -> None:
        if self.model_type not in [AutoMLModel.CLASSIFIER, AutoMLModel.REGRESSOR]:
            raise Exception(f"Model type '{self.model_type}' not supported yet!")

        def log_models(trained_models: List[AutoMLModel]) -> None:
            model_comparison = {model.name: model.score for model in trained_models}
            layer.log({"models": model_comparison})

        def training_func() -> Any:
            # Prepare dataset
            train_dataset = TrainDataset(self.df, self.features, self.target)

            # Run all models matches the requested model type
            best_model = None
            best_score = 0

            trained_models = []
            for automl_model in self.automl_models:
                trainer = automl_model()

                if trainer.model_type == self.model_type:
                    trained_models.append(trainer)
                    trainer.train(train_dataset)
                    if trainer.compare_score(best_score):
                        best_model = trainer
                        best_score = trainer.score

                    # Update models scoreboard after training
                    log_models(trained_models)

            if best_model is None:
                raise Exception("AutoML failed! Check your model type!")

            trained_model = best_model.model

            layer.log({"best model": best_model.name})
            layer.log({"best score": best_score})

            if best_model.feature_importances is not None:
                import matplotlib as plt  # type: ignore

                plt.use("agg")
                feat_importances = pd.DataFrame(
                    best_model.feature_importances, index=self.features, columns=["Importance"]
                )
                feat_importances.sort_values(by="Importance", ascending=False, inplace=True)
                layer.log({"feature importances": feat_importances.plot(kind="bar", figsize=(12, 7))})

            return trained_model

        model_decorator(model_name)(training_func)()  # pylint: disable=no-value-for-parameter
